import pytest
import json
from transmute_core import (
    ParamExtractor, NoArgument,
    describe, annotate, default_context,
    APIException
)
from transmute_core.function import TransmuteFunction


@describe(paths=["/"], query_parameters=["pos"])
@annotate({"pos": str})
def pos_type(pos):
    pass


@describe(paths=["/"])
@annotate({"framework_arg": str})
def with_framework_arg(framework_arg):
    pass


@describe(
    paths=["/"],
    query_parameters=["query"],
    header_parameters=["header"],
    path_parameters=["path"],
    body_parameters=["body"]

)
@annotate({
    "query": str, "header": str, "path": str,
    "body": str
})
def all_param_type(query=1, header=2, path=3, body=4):
    pass


class ParamExtractorMock(ParamExtractor):

    body = json.dumps({"body": "body"})

    def _get_framework_args(self):
        return {
            "framework_arg": "framework_arg"
        }

    def _query_argument(self, key, is_list):
        if is_list:
            return ["query"]
        else:
            return "query"

    def _header_argument(self, key):
        return "header"

    def _path_argument(self, key):
        return "path"


@pytest.fixture
def all_param_type_transmute_func():
    """ ensure retrieval of all parameter types is honored. """
    return TransmuteFunction(all_param_type)


def test_extract_params(all_param_type_transmute_func):

    extractor = ParamExtractorMock()
    args, kwargs = extractor.extract_params(
        default_context, all_param_type_transmute_func,
        "application/json"
    )
    assert args == []
    assert kwargs == {
        "query": "query",
        "header": "header",
        "path": "path",
        "body": "body"
    }


def test_extract_params_no_arguments(all_param_type_transmute_func):
    """ if no arguments are passed, use the defaults """

    extractor = ParamExtractorMock()
    extractor._query_argument = lambda *args: NoArgument
    extractor._header_argument = lambda *args: NoArgument
    extractor._path_argument = lambda *args: NoArgument
    args, kwargs = extractor.extract_params(
        default_context, all_param_type_transmute_func,
        "application/json"
    )
    assert args == []
    assert kwargs == {
        "query": 1,
        "header": 2,
        "path": 3,
        "body": "body"
    }


def test_body_with_only_default_args_can_be_empty(all_param_type_transmute_func):
    """ if no body is passed, and all body args have defaults, it is still a valid request. """

    extractor = ParamExtractorMock()
    extractor.body = ""
    args, kwargs = extractor.extract_params(
        default_context, all_param_type_transmute_func,
        "application/json"
    )
    assert args == []
    assert kwargs == {
        "query": "query",
        "header": "header",
        "path": "path",
        "body": 4,
    }


def test_extract_params_bad_body_content_type(all_param_type_transmute_func):
    """ if no arguments are passed, use the defaults """

    extractor = ParamExtractorMock()
    with pytest.raises(APIException):
        extractor.extract_params(
            default_context, all_param_type_transmute_func,
            "application/myson"
        )


def test_extract_params_positional_args():
    """ if no arguments are passed, use the defaults """
    tf = TransmuteFunction(pos_type)
    extractor = ParamExtractorMock()
    args, kwargs = extractor.extract_params(
        default_context, tf, "application/json"
    )
    assert args == ["query"]
    assert kwargs == {}


def test_with_framework_arg():
    """ """
    tf = TransmuteFunction(with_framework_arg)
    extractor = ParamExtractorMock()
    args, kwargs = extractor.extract_params(
        default_context, tf, "application/json"
    )
    assert args == ["framework_arg"]
    assert kwargs == {}



from swagger_schema import (
    BodyParameter, QueryParameter, HeaderParameter, PathParameter,
)
from ..function.signature import NoDefault


def get_swagger_parameters(parameters, context):
    ret_parameters = []
    for name, details in parameters.query.items():
        ret_parameters.append(QueryParameter({
            "name": name,
            "required": details.default is NoDefault,
            "type": context.serializers.to_json_schema(details.type)["type"],
        }))
    for name, details in parameters.header.items():
        ret_parameters.append(HeaderParameter({
            "name": name,
            "required": details.default is NoDefault,
            "type": context.serializers.to_json_schema(details.type)["type"],
        }))
    for name, details in parameters.body.items():
        ret_parameters.append(BodyParameter({
            "name": name,
            "required": details.default is NoDefault,
            "schema": context.serializers.to_json_schema(details.type),
        }))
    for name, details in parameters.path.items():
        ret_parameters.append(PathParameter({
            "name": name,
            "required": True,
            "type": context.serializers.to_json_schema(details.type)["type"],
        }))
    return ret_parameters

def _extract_base_type(context, details):
    """ extract the basic type of the context """



import json
from .interface import ContentTypeSerializer
from ..exceptions import SerializationException
from ..compat import string_type


class JsonSerializer(ContentTypeSerializer):

    content_type = ["application/json"]

    @staticmethod
    def dump(data):
        """
        should return back a bytes (or string in python 2),
        representation of your object, to be used in e.g. response
        bodies.
        """
        return json.dumps(data).encode("UTF-8")

    @property
    def main_type(self):
        return self.content_type[0]

    @staticmethod
    def load(raw_bytes):
        """
        given a bytes object, should return a base python data
        structure that represents the object.
        """
        try:
            if not isinstance(raw_bytes, string_type):
                raw_bytes = raw_bytes.decode()
            return json.loads(raw_bytes)
        except ValueError as e:
            raise SerializationException(str(e))

    @staticmethod
    def can_handle(content_type_name):
        """
        given a content type, returns true if this serializer
        can convert bodies of the given type.
        """
        return "json" in content_type_name



import pytest
from transmute_core.response_shape import (
    ResponseShapeSimple, ResponseShapeComplex
)
from swagger_schema import Schema


@pytest.fixture
def schema():
    return Schema({"type": "number"})


def test_response_shape_simple(schema):
    assert ResponseShapeSimple.create_body({"result": "foo"}) == "foo"
    assert ResponseShapeSimple.swagger(schema).to_primitive() ==\
        schema.to_primitive()


def test_response_shape_complex(schema):
    assert ResponseShapeComplex.create_body({"result": "foo", "code": 200}) ==\
        {"result": "foo", "code": 200}

    assert ResponseShapeComplex.swagger(schema).to_primitive() == {
        "title": "SuccessObject",
        "type": "object",
        "properties": {
            "result": schema.to_primitive(),
            "success": {"type": "boolean"},
            "code": {"type": "number"}
        },
        "required": ["success", "result", "code"]
    }



from transmute_core.function.attributes import TransmuteAttributes


def test_merge():
    left = TransmuteAttributes(
        methods=["POST"], query_parameters=["a"]
    )
    right = TransmuteAttributes(
        methods=["PUT"], query_parameters=["b"],
        body_parameters=["c"]
    )
    joined = left | right
    assert joined.methods == set(["POST", "PUT"])
    assert joined.query_parameters == set(["a", "b"])
    assert joined.body_parameters == set(["c"])


def test_merge_response_type_by_code():
    left = TransmuteAttributes(response_types={
        200: {"type": bool}, 201: {"type": bool}
    })
    right = TransmuteAttributes(response_types={
        201: {"type": str}
    })
    joined = left | right
    assert joined.response_types == {
        200: {"type": bool}, 201: {"type": str}
    }


def test_merge_response_success_code():
    left = TransmuteAttributes(success_code=200)
    right = TransmuteAttributes(success_code=201)
    joined = left | right
    assert joined.success_code == 201



from swagger_schema import Schema


class ResponseShape(object):
    """
    result shapes define the return format of the
    response.
    """

    @staticmethod
    def create_body(result_dict):
        """
        given the result dict from
        transmute_func, return back the
        response object.
        """
        raise NotImplementedError()

    @staticmethod
    def swagger(result_schema):
        """
        given the schema of the inner
        result object, return back the
        swagger schema representation.
        """
        raise NotImplementedError()


class ResponseShapeSimple(ResponseShape):
    """ return back just the result object. """

    @staticmethod
    def create_body(result_dict):
        return result_dict["result"]

    @staticmethod
    def swagger(result_schema):
        return result_schema


class ResponseShapeComplex(ResponseShape):
    """
    return back an object with the result nested,
    providing a little more context on the result:

    * status code
    * success
    * result
    """

    @staticmethod
    def create_body(result_dict):
        return result_dict

    @staticmethod
    def swagger(result_schema):
        return Schema({
            "title": "SuccessObject",
            "type": "object",
            "properties": {
                "result": result_schema,
                "success": {"type": "boolean"},
                "code": {"type": "number"}
            },
            "required": ["success", "result", "code"]
        })



from collections import namedtuple


class NoDefault(object):

    def __str__(self):
        return "NoDefault"

    def __repr__(self):
        return "NoDefault"

NoDefault = NoDefault()


Argument = namedtuple("Argument", ["name", "default", "type"])


class FunctionSignature(object):

    NoDefault = NoDefault

    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs

    def get_argument(self, key):
        if key in self.kwargs:
            return self.kwargs[key]
        for arg in self.args:
            if arg.name == key:
                return arg

    @staticmethod
    def from_argspec(argspec):
        """
        retrieve a FunctionSignature object
        from the argspec and the annotations passed.
        """
        attributes = (getattr(argspec, "args", []) +
                      getattr(argspec, "keywords", []))
        defaults = argspec.defaults or []

        arguments, keywords = [], {}

        attribute_list = (attributes[:-len(defaults)]
                          if len(defaults) != 0
                          else attributes[:])
        for name in attribute_list:
            if name == "self":
                continue
            typ = argspec.annotations.get(name)
            arguments.append(Argument(name, NoDefault, typ))

        if len(defaults) != 0:
            for name, default in zip(attributes[-len(defaults):], defaults):
                typ = argspec.annotations.get(name)
                keywords[name] = Argument(name, default, typ)

        return FunctionSignature(arguments, keywords)



import inspect
from swagger_schema import (
    Operation, Response, Schema, PathItem
)
from ..compat import getfullargspec
from ..context import default_context
from .attributes import TransmuteAttributes
from .signature import FunctionSignature
from .parameters import get_parameters
from ..swagger.function import get_swagger_parameters
from ..exceptions import (
    InvalidTransmuteDefinition,
)
from ..handler import process_result


class TransmuteFunction(object):
    """
    TransmuteFunctions wrap a function and add metadata,
    allowing transmute frameworks to extract that information for
    their own use (such as web handler generation or automatic
    documentation)
    """

    params = ["error_exceptions"]

    def __init__(self, func, args_not_from_request=None):
        """
        :param list(str) args_not_from_request: a list of arguments that
            transmute_function should not retrieve from the request
            (e.g. from query parameters). This is useful for frameworks
            with pass in an argument to represent the request, or response object.
        """
        # arguments should be the arguments passed into
        # the function
        #
        # the return type should be specified. If nothing is
        # passed in, a NoneType is returned.
        # for a dynamic language like Python, it's not a huge deal:
        # transmute will still json serialize whatever object you pass it.
        # however, the return type is valuable for autodocumentation systems.
        attrs = getattr(func, "transmute", TransmuteAttributes())
        self._validate_attributes(attrs)
        # these are the paths that the transmute function
        # should respond to.
        self.paths = attrs.paths
        self.success_code = attrs.success_code
        argspec = getfullargspec(func)
        self.signature = FunctionSignature.from_argspec(argspec)
        self.response_types = self._parse_response_types(argspec, attrs)
        self.parameters = get_parameters(
            self.signature, attrs, arguments_to_ignore=args_not_from_request
        )
        # description should be a description of the api
        # endpoint, for use in autodocumentation
        self.description = inspect.cleandoc(func.__doc__ or "")
        # summary should be a shortened version of description.
        self.summary = self._summary(self.description)
        # error_exceptions represents the exceptions
        # that should be caught and return an API exception
        self.error_exceptions = tuple(attrs.error_exceptions)
        # these are the http methods that should route to the given function.
        self.methods = attrs.methods
        self.raw_func = func
        # this is to make discovery easier.
        # TODO: make sure this doesn't mess up GC, as it's
        # a cyclic reference.
        if inspect.ismethod(func):
            func.__func__.transmute_func = self
        else:
            func.transmute_func = self

    def __call__(self, *args, **kwargs):
        return self.raw_func(*args, **kwargs)

    def get_response_by_code(self, code):
        """ return the return type, by code """
        return self.response_types.get(code, {}).get("type")

    def get_swagger_path(self, context=default_context):
        operation = self.get_swagger_operation(context)
        path = PathItem()
        for m in self.methods:
            setattr(path, m.lower(), operation)
        return path

    def get_swagger_operation(self, context=default_context):
        """
        get the swagger_schema operation representation.
        """
        consumes = produces = context.contenttype_serializers.keys()
        parameters = get_swagger_parameters(self.parameters, context)
        responses = {
            "400": Response({
                "description": "invalid input received",
                "schema": Schema({
                    "title": "FailureObject",
                    "type": "object",
                    "properties": {
                        "success": {"type": "boolean"},
                        "result": {"type": "string"}
                    },
                    "required": ["success", "result"]
                })
            })
        }
        for code, details in self.response_types.items():
            responses[str(code)] = Response({
                "description": details.get("description"),
                "schema": context.response_shape.swagger(
                    context.serializers.to_json_schema(details["type"])
                )
            })
        return Operation({
            "summary": self.summary,
            "description": self.description,
            "consumes": consumes,
            "produces": produces,
            "parameters": parameters,
            "responses": responses,
            "operationId": self.raw_func.__name__
        })

    def process_result(self, context, result_body, exc, content_type):
        """
        given an result body and an exception object,
        return the appropriate result object,
        or raise an exception.
        """
        return process_result(
            self, context,
            result_body, exc, content_type
        )

    @staticmethod
    def _validate_attributes(attrs):
        """
        validate a TransmuteAttributes contains valid configuration
        that can be converted into a TransmuteFunction.
        """
        if len(attrs.paths) == 0:
            raise InvalidTransmuteDefinition(
                "at least one path is required for a valid set of transmute attribute: {0}".format(str(attrs))
            )

    @staticmethod
    def _parse_response_types(argspec, attrs):
        """
        from the given parameters, return back the response type dictionaries.
        """
        return_type = argspec.annotations.get("return") or None
        response_types = attrs.response_types.copy()
        if return_type or len(response_types) == 0:
            response_types[attrs.success_code] = {
                "type": return_type,
                "description": "success"
            }
        return response_types

    @staticmethod
    def _summary(description):
        for s in description.splitlines():
            if s != "":
                return s
        return ""



from .attributes import TransmuteAttributes
from .signature import FunctionSignature
from .response import Response
from .transmute_function import TransmuteFunction


__all__ = [
    "FunctionSignature",
    "Response",
    "TransmuteAttributes",
    "TransmuteFunction",
]



class TransmuteAttributes(object):

    def __init__(self, paths=None, methods=None,
                 query_parameters=None, body_parameters=None,
                 header_parameters=None, path_parameters=None,
                 error_exceptions=None, response_types=None,
                 success_code=200):
        self.paths = set(paths or [])
        self.methods = set(methods or ["GET"])
        self.success_code = success_code
        self.query_parameters = set(query_parameters or [])
        self.body_parameters = set(body_parameters or [])
        self.header_parameters = set(header_parameters or [])
        self.path_parameters = set(path_parameters or [])
        self.error_exceptions = set(error_exceptions or [])
        self.response_types = response_types or {}

    def __or__(self, other):
        """
        merge values from another transmute function, taking the
        union of the two sets.
        """
        paths = self.paths | other.paths
        methods = self.methods | other.methods
        success_code = other.success_code or self.success_code
        query_parameters = self.query_parameters | other.query_parameters
        body_parameters = self.body_parameters | other.body_parameters
        header_parameters = self.header_parameters | other.header_parameters
        path_parameters = self.path_parameters | other.path_parameters
        error_exceptions = self.error_exceptions | other.error_exceptions
        response_types = self.response_types.copy()
        for k, v in other.response_types.items():
            response_types[k] = v
        return TransmuteAttributes(paths, methods,
                                   query_parameters,
                                   body_parameters,
                                   header_parameters,
                                   path_parameters,
                                   error_exceptions,
                                   response_types,
                                   success_code)

    def __str__(self):
        arg_list = []
        for k in ["paths", "methods",
                  "query_parameters", "body_parameters",
                  "header_parameters", "path_parameters",
                  "error_exceptions", "response_types",
                  "success_code"]:
            arg_list.append("{0}={1}".format(
                k, getattr(self, k)
            ))
        return "<TransmuteAttributes {0}>".format(" ".join(arg_list))



from transmute_core.exceptions import NoSerializerFound


class SerializerSet(object):
    """
    composes multiple serializers, delegating commands to one
    that can handle the desired content type.

    SerializerSet implements a dict-like interface. Retrieving
    serializers is done by get the content type item:

    .. code-block:: python

        serializers["application/json"]
    """

    def __init__(self, serializer_list):
        self.serializers = serializer_list
        assert len(self.serializers) > 0, "at least one serializer should be passed!"

    def _get_serializer_for_type(self, content_type):
        for serializer in self.serializers:
            if serializer.can_handle(content_type):
                return serializer
        raise NoSerializerFound(
            "unable to find serializer for " + content_type + "in: " +
            str(self.serializers)
        )

    def __getitem__(self, content_type):
        """
        from a given content type, returns the appropriate serializer.

        the list of serializers are interated, from first added to last,
        and returns the one who's can_handle returns true.
        """
        return self._get_serializer_for_type(content_type)

    def keys(self):
        """
        return a list of the contetn types this set supports.

        this is not a complete list: serializers can accept more than
        one content type. However, it is a good representation of the
        class of content types supported.
        """
        return_value = []
        for s in self.serializers:
            return_value += s.content_type
        return return_value

    @property
    def default(self):
        return self.serializers[0]



import pytest
from transmute_core.contenttype_serializers.serializer_set import SerializerSet


def test_serializer_set_empty_list():
    with pytest.raises(AssertionError):
        SerializerSet([])



from .object_serializers import SchematicsSerializer
from .contenttype_serializers import get_default_serializer_set
from .response_shape import ResponseShapeSimple


class TransmuteContext(object):
    """
    TransmuteContext contains all of the configuration points for a
    framework based off of transmute.

    It is useful for customizing default behaviour in Transmute, such
    as serialization of additional content types, or using different
    serializers for objects to and from basic data times.
    """

    def __init__(self, serializers=None, contenttype_serializers=None,
                 response_shape=None):
        self.serializers = serializers or SchematicsSerializer()
        self.contenttype_serializers = (
            contenttype_serializers or
            get_default_serializer_set()
        )
        self.response_shape = ResponseShapeSimple

# a global context is provided, if a singleton is sufficient
# or deviations from the defaults are unnescessary
default_context = TransmuteContext()
